"""Pure Python wrapping for InterpAnisotropic class."""

from .._xmsinterp.interpolate import InterpAnisotropic as iAni


class InterpAnisotropic(object):
    """Class for performing anisotropic interpolation."""
    def __init__(self, **kwargs):
        """Constructor.

        Args:
            **kwargs (dict): Generic keyword arguments
        """
        if 'instance' in kwargs:
            self._instance = kwargs['instance']
            return
        self._instance = iAni()

    def __str__(self):
        """Returns a string representation of the InterpAnisotropic."""
        return self._instance.__str__()

    def __repr__(self):
        """Returns a string representation of the InterpAnisotropic."""
        return self._instance.__repr__()

    def __eq__(self, other):
        """Equality operator.

        Args:
            other (InterpAnisotropic): InterpAnisotropic to compare

        Returns:
            bool: True if InterpAnisotropics are equal
        """
        other_instance = getattr(other, '_instance', None)
        if not other_instance or not isinstance(other_instance, iAni):
            return False
        return other_instance == self._instance

    def __ne__(self, other):
        """Equality operator.

        Args:
            other (InterpAnisotropic): InterpAnisotropic to compare

        Returns:
            bool: True if InterpAnisotropics are not equal
        """
        result = self.__eq__(other)
        return not result

    def set_points(self, center_line_points, interpolation_points, pick_closest):
        """Sets points that wil be used for the centerline and interpolation.

        Args:
            center_line_points (iterable): Points that make the centerline
            interpolation_points (iterable): Points that will be used to interpolate from
            pick_closest (bool): If true, only keep one transformed interpolation point (that is closest to the
                                 centerline)
        """
        self._instance.SetPoints(center_line_points, interpolation_points, pick_closest)

    def interp_to_point(self, point):
        """Interpolates to the location specified by a_pt and returns the value.

        Args:
            point (tuple): The location of the interpolation point

        Returns:
            The interpolated value
        """
        return self._instance.InterpToPt(point)

    def interp_to_points(self, points):
        """Interpolates to the locations specified by a_pts and returns the values.

        Args:
            points (iterable): The points to interpolate to

        Returns:
            The interpolated values
        """
        return self._instance.InterpToPts(points)

    def get_interpolation_points(self):
        """Get the interpolation points transformed into (s,n,z) space and scaled.

        Returns:
            The transformed and scaled interpolation points (passed to SetPoints).
        """
        return self._instance.GetInterpolationPts()

    def get_transformed_points(self, points, get_closest):
        """Transform points into (s,n,z) space and then scale.

        Args:
            points (iterable): The points to transform into (s,n,z) space.
            get_closest (bool): True to pick only the closest transformed point.

        Returns:
            The points transformed into (s,n,z) space.
        """
        return self._instance.GetTransformedPts(points, get_closest)

    def set_power(self, power):
        """Sets the exponent for the interpolation.

        By default the class does inverse distance squared weighting but the exponent can be changed to any value.

        Args:
            power (float): The exponent used to compute the point weights
        """
        self._instance.SetPower(power)

    def set_x_scale(self, scale):
        """Set the X Scale.

        Args:
            scale (float): X scale
        """
        self._instance.SetXScale(scale)



"""Pure Python wrapping for InterpLinearExtrapIdw class."""
from .interpolator import Interpolator
from .._xmsinterp.interpolate import InterpLinearExtrapIdw as iLin


class InterpLinearExtrapIdw(Interpolator):
    """Class for performing linear interpolation."""
    nodal_function_types = {
        'constant': 0,
        'gradient_plane': 1,
        'quadratic': 2
    }

    nodal_function_point_search_options = {
        'natural_neighbor': 0,
        'nearest_points': 1,
    }

    def __init__(self, points=None, triangles=None, scalars=None, nodal_function=None,
                 point_search_option="natural_neighbor", number_nearest_points=16, blend_weights=True,
                 progress=None, quadrant_oct=False, weight_quadrant_oct=False,
                 weight_number_nearest_points=16, **kwargs):
        """Constructor.

        Args:
            points (list): The points of the source geometry as x,y,z tuples
            triangles (list): The triangles of the source geometry as point index tuples
            scalars (list): The scalar values at the points
            nodal_function (str): The nodal function to use One of: 'constant', 'gradient_plane', 'quadratic'
            point_search_option (str): The point search option. One of: 'natural_neighbor', 'nearest_points'
            number_nearest_points (int): The number of nearest points to consider
            blend_weights (bool): True if weights should be blended
            progress (Observer): Observer object for providing feedback
            weight_number_nearest_points (int): number of points to include in interpolation weight calculation
            quadrant_oct (bool): True if quadrant search should be used
            weight_quadrant_oct (bool): get nearest points from quadrants for weight calculation
            **kwargs (dict): Generic keyword arguments
        """
        if 'instance' in kwargs:
            self._instance = kwargs['instance']
            return

        if points is None:
            raise ValueError('"points" is a required argument.')
        self._check_points(points)

        if triangles is None or len(triangles) == 0:
            triangles = []
        else:
            self._check_triangles(triangles, len(points))

        if scalars is None or len(scalars) == 0:
            scalars = []
        else:
            self._check_scalars(scalars, len(points))

        self._instance = iLin(points, triangles, scalars)

        if nodal_function is not None:
            self.set_use_natural_neighbor(True, nodal_function, point_search_option, number_nearest_points,
                                          blend_weights, progress)
            self.set_idw_nodal_function(nodal_function, number_nearest_points, quadrant_oct, progress)
        self.set_idw_search_options(nearest_point=weight_number_nearest_points, quadrant_oct_search=weight_quadrant_oct)

        super().__init__(**kwargs)

    def __eq__(self, other):
        """Equality operator.

        Args:
            other (InterpLinearExtrapIdw): InterpLinearExtrapIdw to compare

        Returns:
            bool: True if InterpLinearExtrapIdws are equal
        """
        other_instance = getattr(other, '_instance', None)
        if not other_instance or not isinstance(other_instance, iLin):
            return False
        return other_instance == self._instance

    def __ne__(self, other):
        """Equality operator.

        Args:
            other (InterpLinearExtrapIdw): InterpLinearExtrapIdw to compare

        Returns:
            bool: True if InterpLinearExtrapIdws are not equal
        """
        result = self.__eq__(other)
        return not result

    def __repr__(self):
        """Returns a string representation of the InterpLinearExtrapIdw."""
        return '<InterpLinearExtrapIdw - Point Count: {}, Triangle Count: {}>'.format(
            len(self.points),
            int(len(self.triangles) / 3),
        )

    def __str__(self):
        """Returns a string representation of the InterpLinearExtrapIdw."""
        return '<InterpLinearExtrapIdw - Point Count: {}, Triangle Count: {}>'.format(
            len(self.points),
            int(len(self.triangles) / 3),
        )

    @staticmethod
    def _check_points(points):
        """Verifies that there are at least three points.

        Args:
            points (iterable): Array of points
        """
        if len(points) < 3:
            raise ValueError('"points" must be a list of 3 or more points')

    @staticmethod
    def _check_triangles(triangles, point_length):
        """Verifies that the triangles are valid.

        Args:
            triangles (iterable): Array of the triangles
            point_length (int): Count of all points
        """
        if len(triangles) % 3 != 0:
            raise ValueError('"triangles" must be a list of point indexes divisible by 3 defining each triangle.')
        max_triangles = max(triangles)
        if max_triangles > point_length - 1:
            raise ValueError('point in triangles is out of range: {}'.format(max_triangles))
        min_triangles = min(triangles)
        if min_triangles < 0:
            raise ValueError('point in triangles is out of range: {}'.format(min_triangles))

    @staticmethod
    def _check_scalars(scalars, point_length):
        """Verifies that the scalars are valid.

        Args:
            scalars (iterable): Array of scalars
            point_length (int): Count of all points
        """
        if len(scalars) != point_length:
            raise ValueError("Length of scalars must be equal to the length of points.")

    @staticmethod
    def _check_activity(activity, _length, _type):
        """Verifies that the activity is valid.

        Args:
            activity (iterable): Array of activity
            _length (int): Count of all points
            _type (string): 'points' or 'triangles'
        """
        if len(activity) != _length:
            raise ValueError("Length of activity must be equal to the length of {}.".format(_type))

    @staticmethod
    def _get_nodal_function_type(_str):
        """Gets the nodal function type based on the string.

        Args:
            _str (string): Nodal function type: 'constant', 'gradient_plane' or 'quadratic'

        Returns:
            The int value based on the nodal function type
        """
        nf_type = InterpLinearExtrapIdw.nodal_function_types.get(_str, None)
        if nf_type is None:
            raise ValueError('"nodal_function_type" must be one of {}, not {}'.format(
                ", ".join(InterpLinearExtrapIdw.nodal_function_types.keys()), _str))
        return nf_type

    @staticmethod
    def _get_nodal_function_point_search_options(_str):
        """Gets the nodal function point search option.

        Args:
            _str (string): Search option: 'natural_neighbors' or 'nearest_points'

        Returns:
            The int value based on the nodal function point search option
        """
        nf_type = InterpLinearExtrapIdw.nodal_function_point_search_options.get(_str, None)
        if nf_type is None:
            raise ValueError('"nodal_function_point_search_option" must be one of {}, not {}'.format(
                ", ".join(InterpLinearExtrapIdw.nodal_function_point_search_options.keys()), _str))
        return nf_type

    def set_points_and_triangles(self, points, triangles):
        """Adds the triangles to the class.

        Args:
            points (iterable): Array of point locations.
            triangles (iterable): Array of triangles that references the points array. This array will have a size that
                                  is a multiple of 3. The first 3 locations in array represent the first triangle and
                                  will have indices that correspond to locations in the points array.
        """
        self._instance.SetPtsTris(points, triangles)

    @property
    def points(self):
        """Gets the points."""
        return self._instance.GetPts

    @property
    def triangles(self):
        """Gets the triangles."""
        return self._instance.GetTris

    @property
    def scalars(self):
        """Gets the scalars."""
        return self._instance.GetScalars

    @scalars.setter
    def scalars(self, value):
        """Set the scalars that will be used to interpolate from.

        Args:
            value (iterable): Array of interpolation scalar values.
        """
        self._check_scalars(value, len(self.points))
        self._instance.SetScalars(value)

    def interpolate_to_point(self, point):
        """Use the stored triangles to interpolate to a point.

        Returns extrapolation value if the point is outside the triangles.

        Args:
            point (tuple): Location that is interpolated to.

        Returns:
            Interpolated value at point.
        """
        return self._instance.InterpToPt(point)

    def interpolate_to_points(self, points):
        """Calls InterpToPt in a loop.

        Args:
            points (iterable): Locations of points.

        Returns:
            Interpolated values at points.
        """
        return self._instance.InterpToPts(points)

    @property
    def point_activity(self):
        """Gets the point activity."""
        return self._instance.GetPtActivity

    @point_activity.setter
    def point_activity(self, value):
        """Modifies the activity bitset of the class.

        Args:
            value (iterable): Bitset of the activity of the points
        """
        self._check_activity(value, len(self.points), "points")
        self._instance.SetPtActivity(value)

    @property
    def triangle_activity(self):
        """Gets the triangle activity."""
        return self._instance.GetTriActivity

    @triangle_activity.setter
    def triangle_activity(self, value):
        """Modifies the activity bitset of the class.

        Args:
            value (iterable): Bitset of the activity of the triangles
        """
        self._check_activity(value, len(self.triangles) / 3, "triangles")
        self._instance.SetTriActivity(value)

    @property
    def extrapolation_point_indexes(self):
        """Returns vector of point indexes for points that were outside of all triangles."""
        return self._instance.GetExtrapolationPointIndexes

    def triangle_containing_point(self, point):
        """Find the triangle containing the point.

        Args:
            point (tuple): Location used to find a triangle.

        Returns:
            Index of triangle containing point. If None is returned then no triangle contained the point.
        """
        return self._instance.TriContainingPt(point)

    def triangle_envelopes_containing_point(self, point):
        """Find all triangles whose envelop contains the point.

        Args:
            point (tuple): Location used to find a triangle.

        Returns:
            Iterable of triangle indexes containing the point
        """
        return self._instance.TriEnvelopsContainingPt(point)

    def interpolate_weights(self, point):
        """Use the stored triangles to get interpolation weights for a point.

        Returns False if the point is outside the triangles.

        Args:
            point (tuple): Location that is interpolated to.

        Returns:
            tuple of a bool, array of indexes, and array of weights
        """
        return self._instance.InterpWeights(point)

    @property
    def extrapolation_value(self):
        """Set the constant extrapolation value."""
        return self._instance.GetExtrapVal

    @extrapolation_value.setter
    def extrapolation_value(self, value):
        """Get extrapolation value."""
        self._instance.SetExtrapVal(value)

    def set_truncation(self, maximum, minimum):
        """Set the truncation values for the interpolation and turn on truncation.

        Args:
            maximum (float): The maximum value for truncation.
            minimum (float): The minimum value for truncation.
        """
        if maximum < minimum:
            raise ValueError('maximum must be greater than minimum')
        self._instance.SetTrunc(maximum, minimum)

    @property
    def use_clough_tocher(self):
        """Get the option for using Clough Tocher interpolation."""
        return self._instance.GetUseCloughTocher()

    def set_use_clough_tocher(self, on, progress=None):
        """Set the class to use the Clough Tocher interpolation method.

        This is a legacy feature from GMS. Compare to linear.

        Args:
            on (bool): TIndicate if CT should be used.
            progress (observer): Progress bar to give users feed back on the set up process of CT.
                                 If you have a really large set of triangles this may take some time.
        """
        self._instance.SetUseCloughTocher(on, progress)

    @property
    def use_natural_neighbor(self):
        """Get the option for using Natural Neighbor interpolation."""
        return self._instance.GetUseNatNeigh()

    def set_use_natural_neighbor(self, on=True, nodal_function="constant",
                                 point_search_option="natural_neighbor",
                                 number_nearest_points=16, blend_weights=True,
                                 progress=None):
        """Set the class to use natural neighbor (NN) interpolation.

        Args:
            on (bool): Indicate if NN should be used.
            nodal_function (str): Indicates which nodal function to use.
            point_search_option (str): Indicates options for the nearest points when computing
                                                      the nodal functions.
            number_nearest_points (int): The number of nearest points for nodal function computation.
            blend_weights (bool): Option to use a blending function on the calculated weights.
            progress (observer): Progress bar to give user feedback for generation of the nodal functions.
        """
        nft = self._get_nodal_function_type(nodal_function)
        nfpso = self._get_nodal_function_point_search_options(point_search_option)
        self._instance.SetUseNatNeigh(on, nft, nfpso, number_nearest_points,
                                      blend_weights, progress)

    def set_idw_nodal_function(self, nodal_function="constant", number_nearest_points=16, quadrant_oct=False,
                               progress=None):
        """Sets the type of nodal function as well as options for computing nodal functions.

        Args:
            nodal_function (string): The nodal function methodology:
                                          constant (0), gradient plane (1), quadratic (2).
            number_nearest_points (int): The nearest number of points to use when calculating the nodal functions.
            quadrant_oct (bool): Find the nearest number of points in each quadrant (2d) or octant (3d) when computing
                                 nodal functions.
            progress (float): Progress bar to give user feedback.
        """
        nft = self._get_nodal_function_type(nodal_function)
        self._instance.SetIdwNodalFunction(nft, number_nearest_points, quadrant_oct, progress)

    def set_idw_search_options(self, nearest_point, quadrant_oct_search):
        """Sets the search options for how to find the nearest points to the interpolation point.

        The number of nearest points can be specified as well as whether to find the nearest points in each
        quadrant or octant.

        Args:
            nearest_point (int): The number of nearest points to the interpolation point. These points are used to do
                                 the interpolation.
            quadrant_oct_search (bool): Specifies if the search criterion should find the nearest points in each
                                 quadrant (2d) or octant (3d)
        """
        self._instance.SetIdwSearchOpts(nearest_point, quadrant_oct_search)

    @property
    def truncate_interpolated_values(self):
        """Get the option to truncate interpolated values."""
        return self._instance.GetTruncateInterpolatedValues

    @property
    def truncate_min(self):
        """Get minimum truncation value."""
        return self._instance.GetTruncMin

    @property
    def truncate_max(self):
        """Get maximum truncation value."""
        return self._instance.GetTruncMax

    @property
    def native_neighbor_nodal_function(self):
        """Get the value for the Natural Neighbor nodal function."""
        return self._instance.GetNatNeighNodalFunc

    @property
    def native_neighbor_nodal_function_nearest_points_option(self):
        """Get the option for the Natural Neighbor nodal function nearest points."""
        return self._instance.GetNatNeighNodalFuncNearestPtsOption

    @property
    def native_neighbor_nodal_function_number_nearest_points(self):
        """Get the value for the number of nearest points to use when calculating the nodal function."""
        return self._instance.GetNatNeighNodalFuncNumNearestPts

    @property
    def native_neighbor_blend_weights(self):
        """Get the option for blending weights when using Natural Neighbor."""
        return self._instance.GetNatNeighBlendWeights



"""Convenience methods for using interpolator classes."""
import numpy as np

from xms.interp.interpolate import Interpolator

try:
    import xarray as xr
    xr_enabled = True
except ImportError:
    xr_enabled = False

try:
    import pandas as pd
    pandas_enabled = True
except ImportError:
    pandas_enabled = False


def interpolate_to_points(interpolation_points, points, method, x='x', y='y', z='z', **kwargs):
    """Interpolate values from one set of points to another.

    Args:
        interpolation_points: The points to be used for interpolation
        points: The points to interpolate to
        method: The interpolation method to be used
        x: The x dataset
        y: The y dataset
        z: The z dataset

    Returns:
        The interpolated points
    """
    interp_object = Interpolator.get_interpolator(interpolator_type=method, values=interpolation_points, **kwargs)
    return interp_object.interpolate(points, x=x, y=y, z=z)


def _buffer_range(range_min, range_max, buf_by=0.1):
    """Get the (x, y) range of the data and expand by a percentage.

    Args:
        range_min: The minimum
        range_max: The maximum
        buf_by (float, optional): Percentage to expand, default to 10%

    Returns:
        the minimum expanded by the buffer, and the maximum expanded by the buffer
    """
    buf = (range_max - range_min) * buf_by
    return range_min - buf, range_max + buf


def _generate_x_y_coords(interpolation_points, size, buffer_by=0.1):
    """Generate the x and y coordinates of a grid.

    Args:
        interpolation_points: The interpolation points
        size: The size of the grid
        buffer_by: The buffer percentage

    Returns:
        The generated x and y coordinates

    Warning:
        Currently assuming interp_points is a pandas.DataFrame

    Todo:
        Handle other cases (e.g numpy array, xarray, list of tuples, etc.)
    """
    x_range = _buffer_range(interpolation_points.x.min(), interpolation_points.x.max(), buf_by=buffer_by)
    y_range = _buffer_range(interpolation_points.y.min(), interpolation_points.y.max(), buf_by=buffer_by)

    number_of_x_points, number_of_y_points = size
    x = np.linspace(x_range[0], x_range[1], number_of_x_points)
    y = np.linspace(y_range[0], y_range[1], number_of_y_points)
    return x, y


def interpolate_to_grid(interpolation_points, method, grid=None, size=(100, 100), x='x', y='y', z='z',
                        in_place=False, as_xr_dataarray=False, as_xr_dataset=False, **kwargs):
    """Interpolate values from a set of points to the points of a grid.

    Args:
        interpolation_points: The points to be used for interpolation
        method: Interpolation method to be used
        grid: The pandas dataframe to be used
        size: The size of the grid if no grid is provided
        x: The dataset to be used as x on the grid
        y: The dataset to be used as y on the grid
        z: The dataset to be used as z on the grid
        in_place: Make the changes in place
        as_xr_dataarray: return as xarray data array
        as_xr_dataset: return as xarray dataset

    Returns:
        The new dataset as an xarray data array or data set

    Warning:
        This api call requires pandas
    """
    if grid is not None:
        default_x, default_y = list(grid.coords.items())[:2]
        x_coord, y_coord = grid.coords.get(x, default_x), grid.coords.get(y, default_y)
        size = x_coord.size, y_coord.size
    else:
        x_coord, y_coord = _generate_x_y_coords(interpolation_points, size)

    if not pandas_enabled:
        raise RuntimeError("This api function requires pandas to be installed")

    pts = np.stack([i.flatten() for i in np.meshgrid(x_coord, y_coord)], axis=1)
    pts = pd.DataFrame(columns=['x', 'y'], data=pts)
    interpolate_to_points(interpolation_points=interpolation_points, points=pts, method=method, **kwargs)

    z_values = pts.z.values.reshape((size[1], size[0]))
    data = None

    if xr_enabled:
        data = xr.DataArray(coords={'x': x_coord, 'y': y_coord}, dims=['x', 'y'], data=z_values.T, name=z)

    if in_place and grid is not None:
        if xr_enabled and isinstance(grid, xr.DataArray) or hasattr(grid, 'to_dataset'):
            grid.name = z
            grid.data = z_values

        if xr_enabled and isinstance(grid, xr.Dataset) or hasattr(grid, 'to_array'):
            if data is not None:
                grid[z] = data
            else:
                arr = list(grid.data_vars.values())[0].copy()
                # TODO: not handling if default_x, _y are used or if first array has different coords
                assert x in arr.coords
                assert y in arr.coords
                arr.data = z_values
                grid[z] = arr

        return grid
    elif data is not None:  # i.e. xarray is installed
        if as_xr_dataarray:
            return data
        if as_xr_dataset:
            return data.to_dataset()

    return z_values



"""Pure Python wrapping for InterpLinear class."""
from .interpolator import Interpolator
from .._xmsinterp.interpolate import InterpLinear as iLin


class InterpLinear(Interpolator):
    """Class for performing linear interpolation."""
    nodal_function_types = {
        'constant': 0,
        'gradient_plane': 1,
        'quadratic': 2
    }

    nodal_function_point_search_options = {
        'natural_neighbor': 0,
        'nearest_points': 1,
    }

    def __init__(self, points=None, triangles=None, scalars=None, nodal_function=None,
                 point_search_option="natural_neighbor", number_nearest_points=16, blend_weights=True,
                 progress=None, **kwargs):
        """Constructor.

        Args:
            points (list): The points of the source geometry as x,y,z tuples
            triangles (list): The triangles of the source geometry as point index tuples
            scalars (list): The scalar values at the points
            nodal_function (str): The nodal function to use One of: 'constant', 'gradient_plane', 'quadratic'
            point_search_option (str): The point search option. One of: 'natural_neighbor', 'nearest_points'
            number_nearest_points (int): The number of nearest points to consider
            blend_weights (bool): True if weights should be blended
            progress (Observer): Observer object for providing feedback
            **kwargs (dict): Generic keyword arguments
        """
        if 'instance' in kwargs:
            self._instance = kwargs['instance']
            return

        if points is None:
            raise ValueError('"points" is a required argument.')
        self._check_points(points)

        if triangles is None or len(triangles) == 0:
            triangles = []
        else:
            self._check_triangles(triangles, len(points))

        if scalars is None or len(scalars) == 0:
            scalars = []
        else:
            self._check_scalars(scalars, len(points))

        self._instance = iLin(points, triangles, scalars)

        if nodal_function is not None:
            self.set_use_natural_neighbor(True, nodal_function, point_search_option, number_nearest_points,
                                          blend_weights, progress)

        super().__init__(**kwargs)

    def __eq__(self, other):
        """Equality operator.

        Args:
            other (InterpLinear): InterpLinear to compare

        Returns:
            bool: True if InterpLinears are equal
        """
        other_instance = getattr(other, '_instance', None)
        if not other_instance or not isinstance(other_instance, iLin):
            return False
        return other_instance == self._instance

    def __ne__(self, other):
        """Equality operator.

        Args:
            other (InterpLinear): InterpLinear to compare

        Returns:
            bool: True if InterpLinears are not equal
        """
        result = self.__eq__(other)
        return not result

    def __repr__(self):
        """Returns a string representation of the InterpLinear."""
        return '<InterpLinear - Point Count: {}, Triangle Count: {}>'.format(
            len(self.points),
            int(len(self.triangles) / 3),
        )

    def __str__(self):
        """Returns a string representation of the InterpLinear."""
        return '<InterpLinear - Point Count: {}, Triangle Count: {}>'.format(
            len(self.points),
            int(len(self.triangles) / 3),
        )

    @staticmethod
    def _check_points(points):
        """Verifies that there are at least three points.

        Args:
            points (iterable): Array of points
        """
        if len(points) < 3:
            raise ValueError('"points" must be a list of 3 or more points')

    @staticmethod
    def _check_triangles(triangles, point_length):
        """Verifies that the triangles are valid.

        Args:
            triangles (iterable): Array of the triangles
            point_length (int): Count of all points
        """
        if len(triangles) % 3 != 0:
            raise ValueError('"triangles" must be a list of point indexes divisible by 3 defining each triangle.')
        max_triangles = max(triangles)
        if max_triangles > point_length - 1:
            raise ValueError('point in triangles is out of range: {}'.format(max_triangles))
        min_triangles = min(triangles)
        if min_triangles < 0:
            raise ValueError('point in triangles is out of range: {}'.format(min_triangles))

    @staticmethod
    def _check_scalars(scalars, point_length):
        """Verifies that the scalars are valid.

        Args:
            scalars (iterable): Array of scalars
            point_length (int): Count of all points
        """
        if len(scalars) != point_length:
            raise ValueError("Length of scalars must be equal to the length of points.")

    @staticmethod
    def _check_activity(activity, _length, _type):
        """Verifies that the activity is valid.

        Args:
            activity (iterable): Array of activity
            _length (int): Count of all points
            _type (string): 'points' or 'triangles'
        """
        if len(activity) != _length:
            raise ValueError("Length of activity must be equal to the length of {}.".format(_type))

    @staticmethod
    def _get_nodal_function_type(_str):
        """Gets the nodal function type based on the string.

        Args:
            _str (string): Nodal function type: 'constant', 'gradient_plane' or 'quadratic'

        Returns:
            The int value based on the nodal function type
        """
        nf_type = InterpLinear.nodal_function_types.get(_str, None)
        if nf_type is None:
            raise ValueError('"nodal_function_type" must be one of {}, not {}'.format(
                ", ".join(InterpLinear.nodal_function_types.keys()), _str))
        return nf_type

    @staticmethod
    def _get_nodal_function_point_search_options(_str):
        """Gets the nodal function point search option.

        Args:
            _str (string): Search option: 'natural_neighbors' or 'nearest_points'

        Returns:
            The int value based on the nodal function point search option
        """
        nf_type = InterpLinear.nodal_function_point_search_options.get(_str, None)
        if nf_type is None:
            raise ValueError('"nodal_function_point_search_option" must be one of {}, not {}'.format(
                ", ".join(InterpLinear.nodal_function_point_search_options.keys()), _str))
        return nf_type

    def set_points_and_triangles(self, points, triangles):
        """Adds the triangles to the class.

        Args:
            points (iterable): Array of point locations.
            triangles (iterable): Array of triangles that references the points array. This array will have a size that
                                  is a multiple of 3. The first 3 locations in array represent the first triangle and
                                  will have indices that correspond to locations in the points array.
        """
        self._instance.SetPtsTris(points, triangles)

    @property
    def points(self):
        """Gets the points."""
        return self._instance.GetPts

    @property
    def triangles(self):
        """Gets the triangles."""
        return self._instance.GetTris

    @property
    def scalars(self):
        """Gets the scalars."""
        return self._instance.GetScalars

    @scalars.setter
    def scalars(self, value):
        """Set the scalars that will be used to interpolate from.

        Args:
            value (iterable): Array of interpolation scalar values.
        """
        self._check_scalars(value, len(self.points))
        self._instance.SetScalars(value)

    def interpolate_to_point(self, point):
        """Use the stored triangles to interpolate to a point.

        Returns extrapolation value if the point is outside the triangles.

        Args:
            point (tuple): Location that is interpolated to.

        Returns:
            Interpolated value at point.
        """
        return self._instance.InterpToPt(point)

    def interpolate_to_points(self, points):
        """Calls InterpToPt in a loop.

        Args:
            points (iterable): Locations of points.

        Returns:
            Interpolated values at points.
        """
        return self._instance.InterpToPts(points)

    @property
    def point_activity(self):
        """Gets the point activity."""
        return self._instance.GetPtActivity

    @point_activity.setter
    def point_activity(self, value):
        """Modifies the activity bitset of the class.

        Args:
            value (iterable): Bitset of the activity of the points
        """
        self._check_activity(value, len(self.points), "points")
        self._instance.SetPtActivity(value)

    @property
    def triangle_activity(self):
        """Gets the triangle activity."""
        return self._instance.GetTriActivity

    @triangle_activity.setter
    def triangle_activity(self, value):
        """Modifies the activity bitset of the class.

        Args:
            value (iterable): Bitset of the activity of the triangles
        """
        self._check_activity(value, len(self.triangles) / 3, "triangles")
        self._instance.SetTriActivity(value)

    @property
    def extrapolation_point_indexes(self):
        """Returns vector of point indexes for points that were outside of all triangles."""
        return self._instance.GetExtrapolationPointIndexes

    def triangle_containing_point(self, point):
        """Find the triangle containing the point.

        Args:
            point (tuple): Location used to find a triangle.

        Returns:
            Index of triangle containing point. If None is returned then no triangle contained the point.
        """
        return self._instance.TriContainingPt(point)

    def triangle_envelopes_containing_point(self, point):
        """Find all triangles whose envelop contains the point.

        Args:
            point (tuple): Location used to find a triangle.

        Returns:
            Iterable of triangle indexes containing the point
        """
        return self._instance.TriEnvelopsContainingPt(point)

    def interpolate_weights(self, point):
        """Use the stored triangles to get interpolation weights for a point.

        Returns False if the point is outside the triangles.

        Args:
            point (tuple): Location that is interpolated to.

        Returns:
            tuple of a bool, array of indexes, and array of weights
        """
        return self._instance.InterpWeights(point)

    @property
    def extrapolation_value(self):
        """Set the constant extrapolation value."""
        return self._instance.GetExtrapVal

    @extrapolation_value.setter
    def extrapolation_value(self, value):
        """Get extrapolation value."""
        self._instance.SetExtrapVal(value)

    def set_truncation(self, maximum, minimum):
        """Set the truncation values for the interpolation and turn on truncation.

        Args:
            maximum (float): The maximum value for truncation.
            minimum (float): The minimum value for truncation.
        """
        if maximum < minimum:
            raise ValueError('maximum must be greater than minimum')
        self._instance.SetTrunc(maximum, minimum)

    @property
    def use_clough_tocher(self):
        """Get the option for using Clough Tocher interpolation."""
        return self._instance.GetUseCloughTocher()

    def set_use_clough_tocher(self, on, progress=None):
        """Set the class to use the Clough Tocher interpolation method.

        This is a legacy feature from GMS. Compare to linear.

        Args:
            on (bool): TIndicate if CT should be used.
            progress (observer): Progress bar to give users feed back on the set up process of CT.
                                 If you have a really large set of triangles this may take some time.
        """
        self._instance.SetUseCloughTocher(on, progress)

    @property
    def use_natural_neighbor(self):
        """Get the option for using Natural Neighbor interpolation."""
        return self._instance.GetUseNatNeigh()

    def set_use_natural_neighbor(self, on=True, nodal_function="constant",
                                 point_search_option="natural_neighbor",
                                 number_nearest_points=16, blend_weights=True,
                                 progress=None):
        """Set the class to use natural neighbor (NN) interpolation.

        Args:
            on (bool): Indicate if NN should be used.
            nodal_function (str): Indicates which nodal function to use.
            point_search_option (str): Indicates options for the nearest points when computing
                                                      the nodal functions.
            number_nearest_points (int): The number of nearest points for nodal function computation.
            blend_weights (bool): Option to use a blending function on the calculated weights.
            progress (observer): Progress bar to give user feedback for generation of the nodal functions.
        """
        nft = self._get_nodal_function_type(nodal_function)
        nfpso = self._get_nodal_function_point_search_options(point_search_option)
        self._instance.SetUseNatNeigh(on, nft, nfpso, number_nearest_points,
                                      blend_weights, progress)

    @property
    def truncate_interpolated_values(self):
        """Get the option to truncate interpolated values."""
        return self._instance.GetTruncateInterpolatedValues

    @property
    def truncate_min(self):
        """Get minimum truncation value."""
        return self._instance.GetTruncMin

    @property
    def truncate_max(self):
        """Get maximum truncation value."""
        return self._instance.GetTruncMax

    @property
    def native_neighbor_nodal_function(self):
        """Get the value for the Natural Neighbor nodal function."""
        return self._instance.GetNatNeighNodalFunc

    @property
    def native_neighbor_nodal_function_nearest_points_option(self):
        """Get the option for the Natural Neighbor nodal function nearest points."""
        return self._instance.GetNatNeighNodalFuncNearestPtsOption

    @property
    def native_neighbor_nodal_function_number_nearest_points(self):
        """Get the value for the number of nearest points to use when calculating the nodal function."""
        return self._instance.GetNatNeighNodalFuncNumNearestPts

    @property
    def native_neighbor_blend_weights(self):
        """Get the option for blending weights when using Natural Neighbor."""
        return self._instance.GetNatNeighBlendWeights



"""Initialize the module."""
from . import interp_util  # NOQA: F401
from .interp_anisotropic import InterpAnisotropic  # NOQA: F401
from .interp_idw import InterpIdw  # NOQA: F401
from .interp_linear import InterpLinear  # NOQA: F401
from .interp_linear_extrap_idw import InterpLinearExtrapIdw  # NOQA: F401
from .interpolator import Interpolator  # NOQA: F401



"""InterpUtil methods."""
from .._xmsinterp.interpolate import InterpUtil

__all__ = ['generate_raster_idw']


def generate_raster_idw(pts, min_pt, size, num_rows, num_cols, no_data_val, activity):
    """Returns an array of raster values given mesh point data and raster information.

    Args:
        pts (iterable): The measured depths at point locations.
        min_pt (tuple): The minimum element edge size.
        size (tuple): Size of the raster cells in the X and Y directions.
        num_rows (int): Number of rows in the raster.
        num_cols (int): Number of cols in the raster.
        no_data_val (float): The NODATA value to assign to inactive raster pixels in the result.
        activity (iterable): This is the size of the raster return values and is used to determine whether to
         set raster cells to active and interp values to each cell.

    Returns:
        The interpolated values at the given points.
    """
    return InterpUtil.generate_raster_idw(pts, min_pt, size, num_rows, num_cols, no_data_val, activity)



"""Base class for interpolator class."""
from abc import ABC, abstractmethod
import logging

from xms.interp import interpolate

logger = logging.getLogger('xms.interp')


class Interpolator(ABC):
    """Abstract interpolate base class."""

    def __init__(self, **kwargs):
        """Constructor.

        Args:
            **kwargs (dict): Generic keyword arguments
        """
        self.truncation_max = kwargs.pop('truncation_max', None)
        self.truncation_min = kwargs.pop('truncation_min', None)

        if len(kwargs) > 0:
            logger.warning('The following kwargs are being ignored: {}.'.format(kwargs))

    @abstractmethod
    def __str__(self):
        """Returns a string representation of the interpolator."""
        pass

    @abstractmethod
    def __repr__(self):
        """Returns a string representation of the interpolator."""
        pass

    def __eq__(self, other):
        """Equality operator.

        Args:
            other (Interpolator): Interpolator to compare

        Returns:
            bool: True if Interpolators are equal
        """
        other_instance = getattr(other, '_instance', None)
        if not other_instance or not isinstance(other_instance, self.__class__):
            return False
        return other_instance == self._instance

    def __ne__(self, other):
        """Equality operator.

        Args:
            other (Interpolator): Interpolator to compare

        Returns:
            bool: True if Interpolators are not equal
        """
        result = self.__eq__(other)
        return not result

    @abstractmethod
    def set_truncation(self, **kwargs):
        """Set the truncation values for the interpolation and turn on truncation.

        Args:
            kwargs (dict): Generic keyword arguments
        """
        pass

    @abstractmethod
    def interpolate_to_points(self, **kwargs):
        """Calls InterpToPt in a loop.

        Args:
            kwargs (dict): Generic keyword arguments

        Returns:
            Interpolated values at points.
        """
        pass

    def _configure_truncation(self):
        """Configure the truncation options."""
        if self.truncation_max is not None and self.truncation_min is not None:
            self.set_truncation(maximum=self.truncation_max, minimum=self.truncation_min)

    def interpolate(self, pts, x='x', y='y', z='z', in_place=True):
        """Use the stored triangles to interpolate to points.

        Returns extrapolation value if the point is outside the triangles.

        Args:
            pts (pandas.DataFrame): Location that is interpolated to.
            x (str): Name of the x column
            y (str): Name of the y column
            z (str): Name of the z column
            in_place (bool): True if interpolated values should be updated in pts

        Returns:
            Interpolated value at point.
        """
        pts_is_dataframe = False
        try:
            # Assuming pts is a pandas dataframe
            assert x in pts.columns
            assert y in pts.columns
            values = pts[[x, y]].values
            pts_is_dataframe = True
        except AttributeError:
            # assume pts is numpy array
            values = pts

        self._configure_truncation()
        z_values = self.interpolate_to_points(values)

        if z and pts_is_dataframe and in_place:
            pts[z] = z_values
            return pts

        return z_values

    @classmethod
    def get_interpolator_types(cls, as_dict=False):
        """Get the strings associated with the interpolator types enum.

        Args:
            as_dict (bool): True if return value should have mapping between strings and enum values

        Returns:
            dict or list: The interpolator types
        """
        types = {
            'linear': interpolate.InterpLinear,
            'idw': interpolate.InterpIdw,
            'natural_neighbor': interpolate.InterpLinear,
        }
        if as_dict:
            return types
        return list(types.keys())

    @classmethod
    def get_interpolator(cls, interpolator_type, values, **kwargs):
        """Create a new interpolator.

        Args:
            interpolator_type:
            values (pandas.DataFrame): The points
            **kwargs (dict): Generic keyword arguments

        Returns:
            Interpolator: see description
        """
        types = cls.get_interpolator_types(as_dict=True)

        try:
            # assume values is a pandas dataframe
            values = values.values
        except AttributeError:
            pass

        try:
            return types[interpolator_type](points=values, **kwargs)
        except IndexError:
            raise ValueError('Interpolator kind "{}" is not recognized.'.format(interpolator_type))



"""Pure Python wrapping for InterpLinear class."""
from .interpolator import Interpolator
from .._xmsinterp.interpolate import InterpIdw as iIdw


class InterpIdw(Interpolator):
    """Class for performing IDW interpolation."""
    weights = {
        'classic': 0,
        'modified': 1
    }

    nodal_function_types = {
        'constant': 0,
        'gradient_plane': 1,
        'quadratic': 2
    }

    def __init__(self, points=None, triangles=None, scalars=None,
                 nodal_function=None, number_nearest_points=16,
                 quadrant_oct=False, progress=None, is_2d=True,
                 weight_number_nearest_points=16, weight_quadrant_oct=False,
                 **kwargs):
        """Constructor.

        Args:
            points (list): The points of the source geometry as x,y,z tuples
            triangles (list): The triangles of the source geometry as point index tuples
            scalars (list): The scalar values at the points
            nodal_function (str): The nodal function to use One of: 'constant', 'gradient_plane', 'quadratic'
            number_nearest_points (int): The number of nearest points to consider
            quadrant_oct (bool): True if quadrant search should be used
            progress (Observer): Observer object for providing feedback
            is_2d (bool): flag for 2d vs 3d interpolation
            weight_number_nearest_points (int): number of points to include in interpolation weight calculation
            weight_quadrant_oct (bool): get nearest points from quadrants for weight calculation
            **kwargs (dict): Generic keyword arguments
        """
        if 'instance' in kwargs:
            self._instance = kwargs['instance']
            return

        if points is None:
            raise ValueError('"points" is a required argument.')
        self._check_points(points)

        # triangles are not used by IDW
        # if triangles is None or len(triangles) == 0:
        #     triangles = []
        # else:
        #     self._check_triangles(triangles, len(points))

        if scalars is None or len(scalars) == 0:
            scalars = []
        else:
            self._check_scalars(scalars, len(points))

        self._instance = iIdw([], [], [])
        self._instance.SetPts(points, is_2d)
        self._instance.SetScalars(scalars)

        if nodal_function is not None:
            self.set_nodal_function(nodal_function, number_nearest_points, quadrant_oct, progress)
        self.set_search_options(nearest_point=weight_number_nearest_points, quadrant_oct_search=weight_quadrant_oct)

        super().__init__(**kwargs)

    def __eq__(self, other):
        """Equality operator.

        Args:
            other (InterpIdw): InterpIdw to compare

        Returns:
            bool: True if InterpIdws are equal
        """
        other_instance = getattr(other, '_instance', None)
        if not other_instance or not isinstance(other_instance, iIdw):
            return False
        return other_instance == self._instance

    def __ne__(self, other):
        """Equality operator.

        Args:
            other (InterpIdw): InterpIdw to compare

        Returns:
            bool: True if InterpIdws are not equal
        """
        result = self.__eq__(other)
        return not result

    def __str__(self):
        """Returns a string representation of the InterpIdw."""
        return self._instance.__str__()

    def __repr__(self):
        """Returns a string representation of the InterpIdw."""
        return self._instance.__repr__()

    @staticmethod
    def _check_points(points):
        """Verifies that there are at least three points.

        Args:
            points (iterable): All of the points
        """
        if len(points) < 3:
            raise ValueError('"points" must be a list of 3 or more points')

    @staticmethod
    def _check_triangles(triangles, point_length):
        """Verifies that the triangles are correct.

        Args:
            triangles (iterable): All of the triangles
            point_length (int): Count of all of the points
        """
        if len(triangles) % 3 != 0:
            raise ValueError('"triangles" must be a list of point indexes divisible by 3 defining each triangle.')
        max_triangles = max(triangles)
        if max_triangles > point_length - 1:
            raise ValueError('point in triangles is out of range: {}'.format(max_triangles))
        min_triangles = min(triangles)
        if min_triangles < 0:
            raise ValueError('point in triangles is out of range: {}'.format(min_triangles))

    @staticmethod
    def _check_scalars(scalars, point_length):
        """Verifies that there are the same number of points and scalars.

        Args:
            scalars (iterable): All of the scalars
            point_length (int): Count of all of the points
        """
        if len(scalars) != point_length:
            raise ValueError("Length of scalars must be equal to the length of points.")

    @staticmethod
    def _check_activity(activity, _length, _type):
        """Verifies that the activity length and the length of the points or triangles are the same.

        Args:
            activity (iterable): Activity at a given location
            _length (int): Count of all of the points or triangles
            _type (string): 'points' or 'triangles'
        """
        if len(activity) != _length:
            raise ValueError("Length of activity must be equal to the length of {}.".format(_type))

    @staticmethod
    def _get_weight(_str):
        """Get the specified weight.

        Args:
            _str (string): Key for weights

        Returns:
            int: value from weights
        """
        weight_type = InterpIdw.weights.get(_str, None)
        if weight_type is None:
            raise ValueError('"weights" must be one of {}, not {}'.format(
                ", ".join(InterpIdw.weights.keys()), _str))
        return weight_type

    @staticmethod
    def _get_nodal_function_type(_str):
        """Gets the nodal_function_types int value from key string.

        Args:
            _str (string): Key for nodal_function_types

        Returns:
            int: value from nodal_function_types
        """
        nodal_function_type = InterpIdw.nodal_function_types.get(_str, None)
        if nodal_function_type is None:
            raise ValueError('"nodal_function_type" must be one of {}, not {}'.format(
                ", ".join(InterpIdw.nodal_function_types.keys()), _str))
        return nodal_function_type

    def set_points_and_triangles(self, points, triangles):
        """Sets the points that will be used to do the interpolation.

        Args:
            points (iterable): Array of the point locations
            triangles (iterable): All of the triangles
        """
        self._instance.SetPtsTris(points, triangles)

    @property
    def scalars(self):
        """Gets all of the scalars."""
        return self._instance.GetScalars

    @scalars.setter
    def scalars(self, value):
        """Sets the scalar values that will be used to do the interpolation.

        Args:
            value (iterable): Array of scalars
        """
        self._check_scalars(value, len(self.points))
        self._instance.SetScalars(value)

    @property
    def points(self):
        """Gets all of the points."""
        return self._instance.GetPts

    def set_points(self, points, is_2d):
        """Sets the points that will be used to do the interpolation.

        Args:
            points (iterable): Array of the point locations
            is_2d (bool): Flag if this is 2D
        """
        self._check_points(points)
        self._instance.SetPts(points, is_2d)

    @property
    def triangles(self):
        """Gets all of the triangles."""
        return self._instance.GetTris

    def interpolate_to_point(self, point):
        """Interpolates to the location specified by a_pt and returns the value.

        Args:
            point (tuple): The location of the interpolation point.

        Returns:
            The interpolated value.
        """
        return self._instance.InterpToPt(point)

    def interpolate_to_points(self, points):
        """Interpolates to an array of points and fills in an array of scalars.

        Args:
            points (iterable): Array of points to interpolate to.
        """
        return self._instance.InterpToPts(points)

    @property
    def point_activity(self):
        """Gets the point activity."""
        return self._instance.GetPtActivity

    @point_activity.setter
    def point_activity(self, value):
        """Sets the activity on the point being used to interpolate.

        Args:
            value (iterable): Bitset of point activity.
        """
        self._check_activity(value, len(self.points), "points")
        self._instance.SetPtActivity(value)

    @property
    def triangle_activity(self):
        """Gets the triangle activity."""
        return self._instance.GetTriActivity

    @triangle_activity.setter
    def triangle_activity(self, value):
        """Sets the triangle activity.

        Args:
            value (iterable): Bitset of point activity.
        """
        self._check_activity(value, len(self.triangles) / 3, "triangles")
        self._instance.SetTriActivity(value)

    def set_truncation(self, maximum, minimum):
        """Set the truncation values for the interpolation and turn on truncation.

        Args:
            maximum (float): The maximum value for truncation.
            minimum (float): The minimum value for truncation.
        """
        if maximum < minimum:
            raise ValueError('The truncation maximum must be greater than minimum')
        self._instance.SetTrunc(maximum, minimum)

    @property
    def truncate_interpolation_values(self):
        """Gets the truncation interpolation values."""
        return self._instance.GetTruncateInterpolatedValues

    @property
    def truncate_min(self):
        """Gets the truncation minimum."""
        return self._instance.GetTruncMin

    @property
    def truncate_max(self):
        """Gets the truncation maximum."""
        return self._instance.GetTruncMax

    @property
    def power(self):
        """Gets the power."""
        return self._instance.GetPower

    @power.setter
    def power(self, value):
        """Sets the exponent for the interpolation.

        By default the class does inverse distance squared weighting but the exponent can be changed to any value.

        Args:
            value (float): The exponent used to compute the point weights
        """
        self._instance.SetPower(value)

    @property
    def search_options_number_nearest_points(self):
        """Gets the search options number nearest points."""
        return self._instance.GetSearchOptsNumNearestPts

    @property
    def search_options_use_quadrant_search(self):
        """Gets the search options using quadrant search."""
        return self._instance.GetSearchOptsUseQuadrantSearch

    @property
    def weight_calculation_method(self):
        """Gets the weight calculation method."""
        return self._instance.GetWeightCalcMethod

    @weight_calculation_method.setter
    def weight_calculation_method(self, weight):
        """Sets the method for calculating the weights.

        The classic just uses 1/distance^exponent. The modified method uses another formulation based on the
        distance of the furtherest location from the interpolation pt.

        Args:
            weight (string): 'modified' or 'classic'
        """
        weight_type = self._get_weight(weight)
        self._instance.SetWeightCalcMethod(weight_type)

    @property
    def nodal_function_type(self):
        """Gets the nodal function type."""
        return self._instance.GetNodalFunctionType

    @property
    def nodal_function_number_nearest_points(self):
        """Gets the nodal function number nearest the points."""
        return self._instance.GetNodalFunctionNumNearestPts

    @property
    def nodal_function_use_quadrant_search(self):
        """Gets the nodal function using quadrant search."""
        return self._instance.GetNodalFunctionUseQuadrantSearch

    def set_observer(self, progress=None):
        """Set the observer class so that feedback on the interpolation process can be received.

        Args:
            progress (observer): The observer
        """
        self._instance.SetObserver(progress)

    def set_search_options(self, nearest_point, quadrant_oct_search):
        """Sets the search options for how to find the nearest points to the interpolation point.

        The number of nearest points can be specified as well as whether to find the nearest points in each
        quadrant or octant.

        Args:
            nearest_point (int): The number of nearest points to the interpolation point. These points are used to do
                                 the interpolation.
            quadrant_oct_search (bool): Specifies if the search criterion should find the nearest points in each
                                 quadrant (2d) or octant (3d)
        """
        self._instance.SetSearchOpts(nearest_point, quadrant_oct_search)

    def set_nodal_function(self, nodal_function="constant", number_nearest_points=16, quadrant_oct=False,
                           progress=None):
        """Sets the type of nodal function as well as options for computing nodal functions.

        Args:
            nodal_function (string): The nodal function methodology:
                                          constant (0), gradient plane (1), quadratic (2).
            number_nearest_points (int): The nearest number of points to use when calculating the nodal functions.
            quadrant_oct (bool): Find the nearest number of points in each quadrant (2d) or octant (3d) when computing
                                 nodal functions.
            progress (float): Progress bar to give user feedback.
        """
        nft = self._get_nodal_function_type(nodal_function)
        self._instance.SetNodalFunction(nft, number_nearest_points, quadrant_oct, progress)

    def set_save_weights(self, save):
        """Sets the flag for saving weights.

        Args:
            save (bool): Specifies if the weights should be saved
        """
        self._instance.SetSaveWeights(save)

    def interpolate_weights(self, point):
        """Given a location and an array of points the weights associated with array of points are calculated.

        Args:
            point (tuple): Location of the interpolation point
        """
        return self._instance.InterpWeights(point)

    def set_multithreading(self, use_multithreading):
        """Sets the flag for multi-threading.

        Args:
            use_multithreading (bool): Specifies if multi-threading will be used
        """
        self._instance.SetMultiThreading(use_multithreading)
